{% extends "layout.html" %}
{% block body %}
<p>Blog {% if mode == 'index' %}Index{% else %}Post {{ post_id }}{% endif %}</p>
<p>
  How about going to <a href="{{ req.url_for('index') }}">the index</a>.
</p>
{% endblock %}
